#include <stdio.h>
#include <termios.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdint.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <getopt.h>

#include "sha1.h"

#define BAUDRATE B115200
#define STAGE_2_WRITE_ALL_AT_ONCE 1 /* Write stage 2 in one write() */
#define STAGE_2_WRITE_SIZE 1
#define STAGE_3_WRITE_ALL_AT_ONCE 1 /* Write stage 3 in one write() */
#define STAGE_3_WRITE_SIZE 1
#define FERNLY_USB_LOADER_ADDR 0x7000c000

#define ASSERT(x) do { if ((x)) exit(1); } while(0)

static const uint32_t mtk_config_offset = 0x80000000;
static struct termios old_termios;

/* "general file header", struct gfh_header is actually a lead-in
   for different header types (as specified by type field). */
struct gfh_header {
	uint32_t magic_ver; /* 'MMM', highest byte - version */
	uint16_t size;      /* Total header size, incl. struct gfh_header */
	uint16_t type;      /* 0 - gfh_file_info */
};

struct gfh_file_info {
	struct gfh_header header;
	uint8_t         id[12]; /* "FILE_INFO", zero-padded */
	uint32_t        file_ver;
	uint16_t        file_type;
	uint8_t         flash_dev;
	uint8_t         sig_type;
	uint32_t        load_addr;
	uint32_t        file_len;
	uint32_t        max_size;
	uint32_t        content_offset;
	uint32_t        sig_len;
	uint32_t        jump_offset;
	uint32_t        attr;
};



int print_hex_offset(const void *block, int count, int offset, uint32_t start)
{
	int byte;
	const uint8_t *b = block;
	count += offset;
	b -= offset;
	for ( ; offset < count; offset += 16) {
		printf("%08x", start + offset);

		for (byte = 0; byte < 16; byte++) {
			if (byte == 8)
				printf(" ");
			printf(" ");
			if (offset + byte < count)
				printf("%02x", b[offset + byte] & 0xff);
			else
				printf("  ");
		}

		printf("  |");
		for (byte = 0; byte < 16 && byte + offset < count; byte++)
			printf("%c", isprint(b[offset + byte]) ?
				  b[offset + byte] :
				  '.');
		printf("|\n");
	}
	return 0;
}

int print_hex(const void *block, int count, uint32_t start)
{
	return print_hex_offset(block, count, 0, start);
}

int fernvale_sign_bootloader(int binfd, int boutfd, int32_t addr, uint32_t stack,
			uint32_t unk, const char *filename) {

	struct stat stats;
	struct gfh_file_info *info;
	uint16_t checksum, checksum_calc;
	uint32_t size, shortsize;
	int ret;
	int tmp;

	ret = 0;


	if (-1 == fstat(binfd, &stats)) {
		perror("Unable to get file stats");
		close(binfd);
		return -1;
	}

	size = stats.st_size;

        printf("Filesize: %d\n",size);

	uint8_t bfr[size];

	if (size != read(binfd, bfr, size)) {
		perror("Unable to read file for signing");
		close(binfd);
		return -1;
	}

	close(binfd);

	printf("PATCH BOOTLOADER\n");

	info = (struct gfh_file_info *)((char*)bfr+addr);


	printf("Header-len: %d\n",sizeof(struct gfh_header));

        printf("Info:\n");
        print_hex(bfr+addr,sizeof(struct gfh_file_info),0);

        printf("Id: ");
	int j;
        for (j = 0; j < 12; j++)
                printf("%c", info->id[j]);
        printf("\n");
        printf("Version: %d\n", info->file_ver);
        printf("Type: %d\n", info->file_type);
        printf("Flash device: %d\n", info->flash_dev);
        printf("File size: %d\n", info->file_len);
        printf("Max size: %d\n", info->max_size);
        printf("Signature type: %d\n", info->sig_type);
        printf("Signature length: %d\n", info->sig_len);
        printf("Load address: 0x%08x\n", info->load_addr);
        printf("Content offset: %d\n", info->content_offset);
        printf("Jump offset: %d %X\n", info->jump_offset,info->jump_offset);
        printf("Attributes: %d\n", info->attr);

	printf("Hash:\n");
	print_hex(bfr + size - info->sig_len, info->sig_len, 0);




        printf("sig_type: %d\n",info->sig_type);
        printf("sig_len: %d\n",info->sig_len);
        if(info->sig_len<0) return(-1);
        printf("sig_filelen: %d\n",info->file_len);
        if(info->file_len<0) return(-1);

	//if (info->sig_type == 2) {
		//uint32_t buffer_size = size;

		printf("Patching bootloader signature... \n");

		/* Remove the previous signature length */
		shortsize=size - info->sig_len;

		/* Clear out old checksum */
		memset(bfr + shortsize, 0, info->sig_len);

		/* Calculate SHA1 */
		SHA1Context sha;
		SHA1Reset(&sha);
		SHA1Input(&sha, bfr, shortsize);
		SHA1Result(&sha, bfr + shortsize);

		/* Convert to SHA1 */
		info->sig_type = 1;
		info->sig_len = 32;
		info->file_len = size;
	//}
        

	write(boutfd,bfr,size);

	return ret;
}



static void print_help(const char *name)
{
	printf("Usage: %s [-l logfile] [-s] [serial port] "
			"[stage 1 bootloader] "
			"[[stage 2 bootloader]] "
			"[payload]\n", name);
	printf("\n");
	printf("Arguments:\n");
	printf("    -l [logfile]      Log boot output to the specified file\n");
	printf("    -w                Wait for serial port to appear\n");
	printf("    -s                Enter boot shell\n");
	printf("    -h                Print this help\n");
	printf("\n");
	printf("If you don't want a stage 2 bootloader, you may omit "
		"it, and this program will jump straight to the "
		"payload, loaded at offset 0x%08x.\n",
		FERNLY_USB_LOADER_ADDR);
	printf("\n");
	printf("The boot shell allows you to interact directly with the stage "
	       "2 bootloader.  If you omit -s, then this program will exit "
	       "after loading either the stage 2 bootloader or the payload.\n");
}

int main(int argc, char **argv) {
	int fwfd = -1, outfd = -1 , logfd = -1;
	char *logname = NULL;
	uint32_t ret;

	fwfd = open(argv[1], O_RDONLY);
	if (-1 == fwfd) {
		perror("Unable to open firmware image");
		exit(1);
	}

	outfd = open(argv[2], O_WRONLY|O_CREAT);
	if (-1 == outfd) {
		perror("Unable to open firmware file");
		exit(1);
	}


//        ret=fernvale_sign_bootloader(fwfd,outfd,0,0,0,"");
//        ret=fernvale_sign_bootloader(fwfd,outfd,0x800,0,0,"");
//        ret=fernvale_sign_bootloader(fwfd,outfd,0x217c,0,0,"");
        ret=fernvale_sign_bootloader(fwfd,outfd,0x3400,0,0,"");
//        ret=fernvale_sign_bootloader(fwfd,outfd,0x10000,0,0,"");
//        ret=fernvale_sign_bootloader(fwfd,outfd,0x1c10cc,0,0,"");


/*
00000800  4d 4d 4d 01 38 00 00 00  46 49 4c 45 5f 49 4e 46  |MMM.8...FILE_INF|
00002180  00 4d 4d 4d 46 49 4c 45  5f 49 4e 46 4f 00 00 00  |.MMMFILE_INFO...|
00003400  4d 4d 4d 01 38 00 00 00  46 49 4c 45 5f 49 4e 46  |MMM.8...FILE_INF|
00010000  4d 4d 4d 01 38 00 00 00  46 49 4c 45 5f 49 4e 46  |MMM.8...FILE_INF|
001c10d0  38 00 00 00 46 49 4c 45  5f 49 4e 46 4f 00 00 00  |8...FILE_INFO...|
*/


	close(fwfd);
        close(outfd);
	return 0;
}

